/***************************************************************************
 * sdla_ec.c	WANPIPE(tm) Multiprotocol WAN Link Driver. 
 *				TDM voice hardware echo cancler.
 *
 * Author: 	Alex Feldman   <al.feldman@sangoma.com>
 *
 *
 * Copyright:	(c) 1995-2005 Sangoma Technologies Inc.
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 * ============================================================================
 * Aug 25, 2005	Alex Feldman	Initial version.
 ******************************************************************************
 */
/*******************************************************************************
**			  INCLUDE FILES
*******************************************************************************/
#if (defined __FreeBSD__) | (defined __OpenBSD__)
# include <net/wanpipe_includes.h>
# include <net/wanpipe_debug.h>
# include <net/wanpipe_defines.h>
# include <net/wanpipe_abstr.h>
# include <net/wanpipe_common.h>
# include <net/wanpipe.h>
# include <net/sdla_ec.h>
#elif (defined __WINDOWS__)
# include <wanpipe\csu_dsu.h>
#else
# include <linux/wanpipe_includes.h>
# include <linux/wanpipe_defines.h>
# include <linux/wanpipe.h>
# include <linux/wanpipe.h>
# include <linux/sdla_ec.h>
#endif


/*******************************************************************************
**			  DEFINES AND MACROS
*******************************************************************************/
#define WAN_OCT6100_READ_LIMIT	0x10000

/*****************************************************************************
**			STRUCTURES AND TYPEDEFS
*****************************************************************************/
struct wan_ec_;
typedef struct wan_ec_dev_ {
	char		*name;
	char		devname[WAN_DRVNAME_SZ+1];
	sdla_t		*card;
	struct wan_ec_	*ec;

	WAN_LIST_ENTRY(wan_ec_dev_)	next;
} wan_ec_dev_t;

typedef struct wan_ec_ {
	char		name[WAN_DRVNAME_SZ+1];
	int		usage;
	int		chip_no;

	int		state;
	int		ec_channels_no;	
	int		max_channels;		/* max number of ec channels (security) */
	int		max_ec_channels;	/* total number of ec channels */		
	unsigned int	ulApiInstanceSize;

	WAN_LIST_HEAD(wan_ec_dev_head, wan_ec_dev_)	ec_dev;
	WAN_LIST_ENTRY(wan_ec_)				next;
} wan_ec_t;


/*****************************************************************************
**			   GLOBAL VARIABLES
*****************************************************************************/
static int wan_ec_no = 0;
WAN_LIST_HEAD(wan_ec_head_, wan_ec_) wan_ec_head = 
		WAN_LIST_HEAD_INITIALIZER(wan_ec_head);

/*****************************************************************************
**			  FUNCTION PROTOTYPES
*****************************************************************************/
static int wan_ec_reset(wan_ec_dev_t*, sdla_t*, int);
static int wan_ec_action(void *pcard, int enable, int channel);

extern unsigned char	aft_read_cpld(sdla_t*, unsigned short);
extern int		aft_write_cpld(void*, unsigned short,unsigned char);

/*****************************************************************************
**			  FUNCTION DEFINITIONS
*****************************************************************************/
void *wan_ec_config(void *pcard, int max_channels)
{
	sdla_t		*card = (sdla_t*)pcard;
	wan_ec_t	*ec = NULL;
	wan_ec_dev_t	*ec_dev = NULL, *ec_dev_new = NULL;
	
	WAN_LIST_FOREACH(ec, &wan_ec_head, next){
		WAN_LIST_FOREACH(ec_dev, &ec->ec_dev, next){
			if (ec_dev->card == NULL || ec_dev->card == card){
				DEBUG_EVENT("%s: Internal Error (%s:%d)\n",
						card->devname,
						__FUNCTION__,__LINE__);
				return NULL;
			}
		}
	}
	WAN_LIST_FOREACH(ec, &wan_ec_head, next){
		WAN_LIST_FOREACH(ec_dev, &ec->ec_dev, next){
			if (card->hw_iface.hw_same(ec_dev->card->hw, card->hw)){
				/* Current OCT6100 chip is already in use */
				break;
			}
		}
		if (ec_dev){
			break;
		}
	}
	ec_dev_new = wan_malloc(sizeof(wan_ec_dev_t));
	if (ec_dev_new == NULL){
		DEBUG_EVENT("%s: ERROR: Failed to allocate memory (%s:%d)!\n",
					card->devname,
					__FUNCTION__,__LINE__);
		return NULL;
	}
	memset(ec_dev_new, 0, sizeof(wan_ec_dev_t));
	
	if (ec_dev == NULL){
		/* First device for current Oct6100 chip */
		ec = wan_malloc(sizeof(wan_ec_t));
		if (ec == NULL){
			DEBUG_EVENT("%s: ERROR: Failed to allocate memory (%s:%d)!\n",
						card->devname,
						__FUNCTION__,__LINE__);
			return NULL;
		}
		memset(ec, 0, sizeof(wan_ec_t));
		ec->chip_no		= ++wan_ec_no;
		ec->state		= WAN_OCT6100_STATE_RESET;
		ec->ec_channels_no	= 0;
		ec->max_channels	= 128;	//max_channels;
		sprintf(ec->name, "wp%dec", ec->chip_no);

		WAN_LIST_INIT(&ec->ec_dev);

		WAN_LIST_INSERT_HEAD(&wan_ec_head, ec, next);
	}else{
		ec = ec_dev->ec;
	}
	ec->usage++;
	ec_dev_new->ec		= ec;
	ec_dev_new->name	= ec->name;
	ec_dev_new->card	= card;
	/* Initialize hwec_bypass pointer */
	card->wandev.ec_iface.action	= wan_ec_action;
	card->wandev.ec_map		= 0;
	memcpy(ec_dev_new->devname, card->devname, sizeof(card->devname));
	WAN_LIST_INSERT_HEAD(&ec->ec_dev, ec_dev_new, next);

	DEBUG_EVENT("%s: Register EC interface %s (usage %d, max ec channels %d)!\n",
					ec_dev_new->devname,
					ec->name,
					ec->usage,
					ec->max_channels);
	return (void*)ec_dev_new;
}

int wan_ec_remove(void *arg, void *pcard)
{
	sdla_t		*card = (sdla_t*)pcard;
	wan_ec_t	*ec = NULL;
	wan_ec_dev_t	*ec_dev = (wan_ec_dev_t*)arg;

	WAN_ASSERT(ec_dev == NULL);
	ec = ec_dev->ec;
	DEBUG_EVENT("%s: Unregister EC interface %s (chip id %d, usage %d)!\n",
					card->devname,
					ec->name,
					ec->chip_no,
					ec->usage);
	ec_dev->card = NULL;
	ec_dev->ec = NULL;
	ec->usage--;
	if (WAN_LIST_FIRST(&ec->ec_dev) == ec_dev){
		WAN_LIST_FIRST(&ec->ec_dev) =
				WAN_LIST_NEXT(ec_dev, next);
		WAN_LIST_NEXT(ec_dev, next) = NULL;
	}else{
		WAN_LIST_REMOVE(ec_dev, next);
	}
	wan_free(ec_dev);
	if (!ec->usage){
		if (WAN_LIST_FIRST(&wan_ec_head) == ec){
			WAN_LIST_FIRST(&wan_ec_head) =
					WAN_LIST_NEXT(ec, next);
			WAN_LIST_NEXT(ec, next) = NULL;
		}else{
			WAN_LIST_REMOVE(ec, next);
		}
		wan_free(ec);
	}
	return 0;
}

static wan_ec_dev_t *wan_ec_search(char *devname)
{
	wan_ec_t	*ec;
	wan_ec_dev_t	*ec_dev = NULL;

	WAN_LIST_FOREACH(ec, &wan_ec_head, next){
		WAN_LIST_FOREACH(ec_dev, &ec->ec_dev, next){
			if (strcmp(ec_dev->devname, devname) == 0){
				return ec_dev;
			}
		}
	}
	return NULL;
}

static int
wan_ec_reset(wan_ec_dev_t *ec_dev, sdla_t *card, int cmd)
{
	wan_ec_t	*ec;
	int		err = -EINVAL;

	WAN_ASSERT(ec_dev == NULL);
	ec = ec_dev->ec;

	if (card->wandev.hwec_reset){
		err = card->wandev.hwec_reset(
					card,
					(cmd==WAN_OCT6100_CMD_SET_RESET) ?
						AFT_HWEC_SET_RESET :
						AFT_HWEC_CLEAR_RESET);
		if (!err){
			if (cmd == WAN_OCT6100_CMD_SET_RESET){
				ec->state = WAN_OCT6100_STATE_RESET;
			}else{
				ec->state = WAN_OCT6100_STATE_READY;
			}
		}
	}
	return err;
}

static int wan_ec_action(void *pcard, int enable, int channel)
{
	sdla_t		*card = (sdla_t*)pcard;
	wan_ec_t	*ec = NULL;
	wan_ec_dev_t	*ec_dev = NULL;
	int		err = -ENODEV;

	ec_dev = wan_ec_search(card->devname);
	WAN_ASSERT(ec_dev == NULL);
	WAN_ASSERT(ec_dev->ec == NULL);
	ec = ec_dev->ec;	

	DEBUG_HWEC("[HWEC]: %s: %s bypass mode for channel %d!\n",
			card->devname,
			(enable) ? "Enable" : "Disable",
			channel);

	if (card->wandev.hwec_action){
		if (enable){
			if (ec->ec_channels_no >= ec->max_channels){
				DEBUG_EVENT("%s: Exceeded maximum number of Echo Canceller channels (max=%d)!\n",
						ec->name,
						ec->max_channels);
				return -ENODEV;
			}
		}
		err = card->wandev.hwec_action(card, enable, channel);
		if (!err){
			if (enable){
				ec->ec_channels_no++;
			}else{
				ec->ec_channels_no--;
			}
		}
	}
	return err;
}

static u32 convert_addr(u32 addr)
{
	addr &= 0xFFFF;
	switch(addr){
	case 0x0000:
		return 0x60;
	case 0x0002:
		return 0x64;
	case 0x0004:
		return 0x68;
	case 0x0008:
		return 0x70;
	case 0x000A:
		return 0x74;
	}
	return 0x00;
}

/*===========================================================================*\
  ReadInternalDword
\*===========================================================================*/
static int
wan_ec_read_internal_dword(wan_ec_dev_t *ec_dev, u32 addr, u32 *data)
{
	sdla_t	*card = ec_dev->card;

	WAN_ASSERT(card == NULL);
	addr = convert_addr(addr);
	if (addr){
		int err;
		err = card->hw_iface.bus_read_4(card->hw, addr, data);
		WP_DELAY(5);
#if 0
		DEBUG_EVENT("%s: Reading Addr=%08X Data=%08X (err=%d)\n",
				__FUNCTION__,addr,*data, err);
#endif
		return err;
	}
	return -EINVAL;
}


/*===========================================================================*\
  WriteInternalDword
\*===========================================================================*/
static int
wan_ec_write_internal_dword(wan_ec_dev_t *ec_dev, u32 addr, u32 data)
{
	sdla_t	*card = ec_dev->card;

	WAN_ASSERT(card == NULL);
	addr = convert_addr(addr);
	if (addr){
		int	err;
#if 0
		DEBUG_EVENT("%s: Writting Addr=%08X Data=%08X\n",
				__FUNCTION__,addr,data);
#endif
		err = card->hw_iface.bus_write_4(card->hw, addr, data);	
		WP_DELAY(5);
		return err;
	}
	return -EINVAL;
}

/*===========================================================================*\
  Oct6100Read
\*===========================================================================*/

static int
wan_ec_read(wan_ec_dev_t *ec_dev, u32 read_addr, u16 *read_data)
{
	u32	data;
	wan_ec_read_internal_dword(ec_dev, read_addr, &data);
	*read_data = (u16)(data & 0xFFFF);
	return 0;
}

/*===========================================================================*\
  Oct6100Write
\*===========================================================================*/

static int
wan_ec_write(wan_ec_dev_t *ec_dev, u32 write_addr, u32 write_data)
{
	return wan_ec_write_internal_dword(ec_dev, write_addr, write_data);
}


/*===========================================================================*\
  Oct6100WriteSequenceSub
\*===========================================================================*/
static u32
wan_ec_write_seq(wan_ec_dev_t *ec_dev, u32 write_addr, u16 write_data)
{
	u32	ulResult;
	u32	ulData;
	u16	usData;
	u32	ulWordAddress;
	u32	i;

	/* Create the word address from the provided byte address. */
	ulWordAddress = write_addr >> 1;

	/* 16-bit indirect access. */

	/* First write to the address indirection registers. */
	ulData = ( ulWordAddress >> 19 ) & 0x1FFF;
	ulResult = wan_ec_write( ec_dev, 0x8, ulData );
	if (ulResult) 
		return ulResult;

	ulData = ( ulWordAddress >> 3 ) & 0xFFFF;
	ulResult = wan_ec_write( ec_dev, 0xA, ulData );
	if (ulResult) 
		return ulResult;

	/* Next, write data word to read/write data registers. */
	ulData = write_data & 0xFFFF;
	ulResult = wan_ec_write( ec_dev, 0x4, ulData );
	if ( ulResult ) 
		return ulResult;


	/* Write the parities and write enables, as well as last three bits
	** of wadd and request the write access. */
	ulData = ( ( 0x0 & 0x3 ) << 14 ) | ( ( 0x3 & 0x3 ) << 12 ) | ( ( ulWordAddress & 0x7 ) << 9 ) | 0x0100;
	ulResult = wan_ec_write( ec_dev, 0x0, ulData );
	if ( ulResult ) 
		return ulResult;

	/* Keep polling register contol0 for the access_req bit to go low. */
	for ( i = 0; i < WAN_OCT6100_READ_LIMIT; i++ )
	{
		ulResult = wan_ec_read( ec_dev, 0, &usData );
		if ( ulResult ) 
			return ulResult;

		if ( ( ( usData >> 8 ) & 0x1 ) == 0x0 ) 
			break;
	}

	if ( i == WAN_OCT6100_READ_LIMIT ){
		DEBUG_EVENT("%s: EC write command reached limit!\n",
				ec_dev->name);
		return WAN_OCT6100_RC_CPU_INTERFACE_NO_RESPONSE;
	}
	return 0;
}


/*===========================================================================*\
  HandleReqWriteOct6100
\*===========================================================================*/
static int
wan_ec_req_write(wan_ec_dev_t *ec_dev, wan_ec_api_t *ec_api)
{
	u32	write_addr = ec_api->u_oct6100_write.ulWriteAddress;
	u16	write_data = ec_api->u_oct6100_write.usWriteData;
	u32	ulResult;

	DEBUG_TEST("%s: EC WRITE API addr=%X data=%X\n",
				ec_dev->ec->name, write_addr, write_data);
	ulResult = wan_ec_write_seq(ec_dev, write_addr, write_data);
	if (ulResult){
		DEBUG_EVENT("%s: Failed to write %04X to addr %08X\n",
						ec_dev->name,
						write_addr,
						write_data);
	}
	ec_api->ulResult = ulResult;
	return 0;
}


/*===========================================================================*\
  HandleReqWriteSmearOct6100
\*===========================================================================*/
static int
wan_ec_req_write_smear(wan_ec_dev_t *ec_dev, wan_ec_api_t *ec_api)
{
	u32	addr = ec_api->u_oct6100_write_swear.ulWriteAddress;
	u32	i, len = ec_api->u_oct6100_write_swear.ulWriteLength;
	u16	data = ec_api->u_oct6100_write_swear.usWriteData;
	u32	ulResult = WAN_OCT6100_RC_OK;

	WAN_ASSERT(ec_dev == NULL);
	for ( i = 0; i < len; i++ ){
		ulResult = wan_ec_write_seq(ec_dev, addr + (i*2), data);
		if (ulResult){
			DEBUG_EVENT("%s: Failed to write %04X to addr %08X\n",
						ec_dev->name,
						addr + (i*2),
						data);
			break;
		}
	}
	ec_api->ulResult = ulResult;
	return 0;
}


/*===========================================================================*\
  HandleReqWriteBurstOct6100
\*===========================================================================*/
static int
wan_ec_req_write_burst(wan_ec_dev_t *ec_dev, wan_ec_api_t *ec_api)
{
	u16	*user_space = ec_api->u_oct6100_write_burst.pWriteData,
		*data = NULL;
	u32	addr = ec_api->u_oct6100_write_burst.ulWriteAddress;
	u32	i, len = ec_api->u_oct6100_write_burst.ulWriteLength;
	u32	ulResult = WAN_OCT6100_RC_OK;
	int	err = 0;

	WAN_ASSERT(ec_dev == NULL);
	data = wan_malloc(len * sizeof(u16));
	if (data == NULL){
		DEBUG_EVENT("%s: Failed to allocate memory space [%s:%d]!\n",
				ec_dev->name,
				__FUNCTION__,__LINE__);
		return -EINVAL;
	}
	err = WAN_COPY_FROM_USER(data, user_space, len * sizeof(u16));
	if (err){
		DEBUG_EVENT("%s: Failed to copy data from user space [%s:%d]!\n",
				ec_dev->ec->name,
				__FUNCTION__,__LINE__);
		wan_free(data);
		return -EINVAL;
	}

	for ( i = 0; i < len; i++ ){
		ulResult = wan_ec_write_seq(ec_dev, addr + (i * 2), data[i]);
		if (ulResult){
			DEBUG_EVENT("%s: Failed to write %04X to addr %08X\n",
						ec_dev->name,
						addr + (i*2),
						data[i]);
			break;
		}
	}
	ec_api->ulResult = ulResult;
	if (data) wan_free(data);
	return 0;
}


/*===========================================================================*\
  Oct6100ReadSequenceSub
\*===========================================================================*/
static u32
wan_ec_read_seq(wan_ec_dev_t *ec_dev, u32 read_addr, u16 *read_data, u32 read_len)
{
	u32	ulResult;
	u32	ulData;
	u32	ulWordAddress;
	u32	ulReadBurstLength;
	u16	usData;
	u32	i;

	/* Create the word address from the provided byte address. */
	ulWordAddress = read_addr >> 1;

	/* Indirect accesses. */

	/* First write to the address indirection registers. */
	ulData = ( ulWordAddress >> 19 ) & 0x1FFF;
	ulResult = wan_ec_write( ec_dev, 0x8, ulData );
	if (ulResult) 
		return ulResult;

	ulData = ( ulWordAddress >> 3 ) & 0xFFFF;
	ulResult = wan_ec_write( ec_dev, 0xA, ulData );
	if (ulResult)
		return ulResult;

	/* Request access. */
	if ( read_len >= 128 )
	{
		ulData = 0x100 | ( ( ulWordAddress & 0x7 ) << 9);
		ulReadBurstLength = 0;
	}
	else
	{
		ulData = 0x100 | ( ( ulWordAddress & 0x7 ) << 9) | read_len;
		ulReadBurstLength = read_len;
	}
	ulResult = wan_ec_write( ec_dev, 0x0, ulData );
	if (ulResult)
		return ulResult;

	/* Keep polling register contol0 for the access_req bit to go low. */
	for ( i = 0; i < WAN_OCT6100_READ_LIMIT; i++ )
	{
		ulResult = wan_ec_read( ec_dev, 0x0, &usData );
		if (ulResult) 
			return ulResult;

		if ( ( ( usData >> 8 ) & 0x1 ) == 0x0 )
			break;
	}
	if ( i == WAN_OCT6100_READ_LIMIT ){
		DEBUG_EVENT("%s: EC read command reached limit!\n",
				ec_dev->name);
		return WAN_OCT6100_RC_CPU_INTERFACE_NO_RESPONSE;
	}

	if ( ( usData & 0xFF ) == 0x1 )
	{
		i = 0;
	}

	/* Retrieve read data. */
	ulResult = wan_ec_read( ec_dev, 0x4, &usData );
	if (ulResult)
		return ulResult;

	if ( ( usData & 0xFF ) == 0x1 )
	{
		i = 0;
	}

	*read_data = usData;
	return 0;
}


static int
wan_ec_req_read(wan_ec_dev_t *ec_dev, wan_ec_api_t *ec_api)
{
	u32	read_addr = ec_api->u_oct6100_read.ulReadAddress;
	u32	ulResult;

	WAN_ASSERT(ec_dev == NULL);
	DEBUG_TEST("%s: EC READ API addr=%X data=????\n",
			ec_dev->ec->name,
			read_addr);
	ulResult = wan_ec_read_seq(
				ec_dev,
				read_addr,
				&ec_api->u_oct6100_read.usReadData,
				1);
	ec_api->ulResult = ulResult;
	if (ulResult){
		DEBUG_EVENT("%s: Failed to read data from addr %08X\n",
					ec_dev->name,
					read_addr);
		return -EINVAL;
	}
	DEBUG_TEST("%s: EC READ API addr=%X data=%X\n",
			ec_dev->ec->name,
			read_addr,
			ec_api->u_oct6100_read.usReadData);
	return 0;
}


static int
wan_ec_req_read_burst(wan_ec_dev_t *ec_dev, wan_ec_api_t *ec_api)
{
	u32	read_addr = ec_api->u_oct6100_read_burst.ulReadAddress,
		read_len = ec_api->u_oct6100_read_burst.ulReadLength,
		i;
	u16	*user_space = ec_api->u_oct6100_read_burst.pReadData,
		*data = NULL, read_data;
	u32	ulResult = WAN_OCT6100_RC_OK;
	int	err;

	WAN_ASSERT(ec_dev == NULL);
	data = wan_malloc(read_len * sizeof(u16));
	if (data == NULL){
		DEBUG_EVENT("%s: Failed allocate memory (%d bytes)\n",
				ec_dev->name,
				read_len * sizeof(u16));
		return -EINVAL;
	}

	for ( i = 0; i < read_len; i++ ){
		ulResult = wan_ec_read_seq(ec_dev, read_addr, &read_data, 1);
		if (ulResult){
			DEBUG_EVENT("%s: Failed to read from addr %X\n",
						ec_dev->name,
						read_addr);
			break;
		}
		data[i]  = (u16)read_data;
		read_addr += 2;
	}
	ec_api->ulResult = ulResult;
	if (ulResult){
		if (data) wan_free(data);
		return -EINVAL;
	}
	err = WAN_COPY_TO_USER(	user_space,
				data,
				read_len * sizeof(u16));
	if (err){
		DEBUG_EVENT("%s: Failed to copy data to user space [%s:%d]!\n",
				ec_dev->ec->name,
				__FUNCTION__,__LINE__);
		if (data) wan_free(data);
		return -EINVAL;
	}
	if (data) wan_free(data);
	return 0;
}

int wan_ec_ioctl(void *arg, struct ifreq *ifr, void *pcard)
{
	sdla_t		*card = (sdla_t*)pcard;
	wan_ec_t	*ec = NULL;
	wan_ec_dev_t	*ec_dev = NULL;
	wan_ec_api_t	*ec_api = NULL;
	int		err = 0;

	WAN_ASSERT(arg == NULL);
	ec_dev = (wan_ec_dev_t*)arg;
	ec_api = wan_malloc(sizeof(wan_ec_api_t));
	if (ec_api == NULL){
		DEBUG_EVENT("%s: Failed allocate memory (%d) [%s:%d]!\n",
				ec_dev->ec->name,
				sizeof(wan_ec_api_t),
				__FUNCTION__,__LINE__);
		return -EINVAL;
	}
	err = WAN_COPY_FROM_USER(
				ec_api,
				ifr->ifr_data,
				sizeof(wan_ec_api_t));
	if (err){
		DEBUG_EVENT("%s: Failed to copy data from user space [%s:%d]!\n",
				ec_dev->ec->name,
				__FUNCTION__,__LINE__);
		wan_free(ec_api);
		return -EINVAL;
	}

	if (ec_api->cmd == WAN_OCT6100_CMD_GET_INFO){
		if (strcmp(card->devname, ec_api->devname)){
			DEBUG_EVENT("%s:%s: Failed to find %s device!\n",
					card->devname, ec_api->if_name,
					ec_api->devname);
			wan_free(ec_api);
			return -EINVAL;
		}
	}
	ec_dev = wan_ec_search(ec_api->devname);
	if (ec_dev == NULL || ec_dev->ec == NULL || ec_dev->card == NULL){
		DEBUG_EVENT("%s: Failed to find %s device [%s:%d]!\n",
				card->devname,
				ec_api->devname,
				__FUNCTION__,__LINE__);
		wan_free(ec_api);
		return -EINVAL;
	}
	card = ec_dev->card;
	ec = ec_dev->ec;

	switch(ec_api->cmd){
	case WAN_OCT6100_CMD_GET_INFO:
		memcpy(ec_api->name, ec->name, strlen(ec->name));
		ec_api->u_ecd.state		= ec->state;
		ec_api->u_ecd.chip_no		= ec->chip_no;
		ec_api->u_ecd.fe_media		= WAN_FE_MEDIA(&card->fe);
		ec_api->u_ecd.fe_lineno		= WAN_FE_LINENO(&card->fe);
		ec_api->u_ecd.fe_max_channels	= WAN_FE_MAX_CHANNELS(&card->fe);
		ec_api->u_ecd.fe_tdmv_law	= WAN_FE_TDMV_LAW(&card->fe);
		ec_api->u_ecd.max_channels	= ec->max_channels;
		ec_api->u_ecd.max_ec_channels	= ec->max_ec_channels;
		ec_api->u_ecd.ulApiInstanceSize = ec->ulApiInstanceSize;
		break;

	case WAN_OCT6100_CMD_SET_INFO:
		ec->state		= ec_api->u_ecd.state;
		ec->chip_no		= ec_api->u_ecd.chip_no;
		ec->max_channels	= ec_api->u_ecd.max_channels;
		ec->max_ec_channels	= ec_api->u_ecd.max_ec_channels;
		ec->ulApiInstanceSize	= ec_api->u_ecd.ulApiInstanceSize;
		break;
	
	case WAN_OCT6100_CMD_CLEAR_RESET:
	case WAN_OCT6100_CMD_SET_RESET:
		err = wan_ec_reset(ec_dev, card, ec_api->cmd);
		break;

	case WAN_OCT6100_CMD_BYPASS_ENABLE:
		err = wan_ec_action(card, 1, ec_api->u_ecd.channel);
		break;

	case WAN_OCT6100_CMD_BYPASS_DISABLE:
		err = wan_ec_action(card, 0, ec_api->u_ecd.channel);
		break;

	case WAN_OCT6100_CMD_API_WRITE:
		err = wan_ec_req_write(ec_dev, ec_api);
		break;

	case WAN_OCT6100_CMD_API_WRITE_SMEAR:
		err = wan_ec_req_write_smear(ec_dev, ec_api);
		break;

	case WAN_OCT6100_CMD_API_WRITE_BURST:
		err = wan_ec_req_write_burst(ec_dev, ec_api);
		break;

	case WAN_OCT6100_CMD_API_READ:
		err = wan_ec_req_read(ec_dev, ec_api);
		break;

	case WAN_OCT6100_CMD_API_READ_BURST:
		err = wan_ec_req_read_burst(ec_dev, ec_api);
		break;

	default:
		DEBUG_EVENT("%s: Unsupported EC API command for %s (%X)\n",
				card->devname, ec->name, ec_api->cmd);
		err = -EINVAL;
		break;
	}

	err = WAN_COPY_TO_USER(
				ifr->ifr_data,
				ec_api,
				sizeof(wan_ec_api_t));
	if (err){
		DEBUG_EVENT("%s: Failed to copy data to user space [%s:%d]!\n",
				ec_dev->ec->name,
				__FUNCTION__,__LINE__);
		wan_free(ec_api);
		return -EINVAL;
	}
	wan_free(ec_api);
	return 0;
}

